/*
 * ----------------------------------------------------------------------------
 *  (C) Copyright Testing Technologies, 2001-2007.  All Rights Reserved.
 *
 *  All copies of this program, whether in whole or in part, and whether
 *  modified or not, must display this and all other embedded copyright
 *  and ownership notices in full.
 *
 *  See the file COPYRIGHT for details of redistribution and use.
 *
 *  You should have received a copy of the COPYRIGHT file along with
 *  this file; if not, write to the Testing Technologies,
 *  Rosenthaler Str. 13, 10119 Berlin, Germany.
 *
 *  TESTING TECHNOLOGIES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *  SOFTWARE. IN NO EVENT SHALL TESTING TECHNOLOGIES BE LIABLE FOR ANY
 *  SPECIAL, DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 *  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 *  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 *  THIS SOFTWARE.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 *  EITHER EXPRESSED OR IMPLIED, INCLUDING ANY KIND OF IMPLIED OR
 *  EXPRESSED WARRANTY OF NON-INFRINGEMENT OR THE IMPLIED WARRANTIES
 *  OF MECHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * ----------------------------------------------------------------------------
 *
 *  AUTHOR:      Dirk Tepelmann
 *  DATE:        January 2005
 *
 *  REVISION INFO:
 *     $Revision: 1.7 $  $Date: 2005/09/13 08:47:39 $
 *     $Source: /net/ttech/cvs_root/IPv6_ETSI_STF276/javasrc/com/testingtech/ttcn/tri/IPv6TestAdapter.java,v $
 *
 * ----------------------------------------------------------------------------
 */
package com.testingtech.ttcn.tri;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.math.BigInteger;
import java.net.Socket;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import jpcap.EthernetPacket;
import jpcap.Jpcap;
import jpcap.JpcapHandler;
import jpcap.JpcapSender;
import jpcap.Packet;

import org.etsi.ttcn.tci.BooleanValue;
import org.etsi.ttcn.tci.CharstringValue;
import org.etsi.ttcn.tci.EnumeratedValue;
import org.etsi.ttcn.tci.OctetstringValue;
import org.etsi.ttcn.tci.RecordOfValue;
import org.etsi.ttcn.tci.RecordValue;
import org.etsi.ttcn.tci.TciCDProvided;
import org.etsi.ttcn.tci.TciTMProvided;
import org.etsi.ttcn.tci.Value;
import org.etsi.ttcn.tri.TriActionTemplate;
import org.etsi.ttcn.tri.TriAddress;
import org.etsi.ttcn.tri.TriCommunicationSA;
import org.etsi.ttcn.tri.TriComponentId;
import org.etsi.ttcn.tri.TriFunctionId;
import org.etsi.ttcn.tri.TriMessage;
import org.etsi.ttcn.tri.TriParameter;
import org.etsi.ttcn.tri.TriParameterList;
import org.etsi.ttcn.tri.TriPlatformPA;
import org.etsi.ttcn.tri.TriPortId;
import org.etsi.ttcn.tri.TriPortIdList;
import org.etsi.ttcn.tri.TriStatus;
import org.etsi.ttcn.tri.TriTestCaseId;

import com.testingtech.generated.codec.LibIpv6_CommonRfcs_TypesAndValues.TabularLibIpv6_CommonRfcs_TypesAndValues;
import com.testingtech.generated.codec.LibIpv6_Interface_TypesAndValues.TabularLibIpv6_Interface_TypesAndValues;
import com.testingtech.generated.codec.LibIpv6_Rfc0768Udp_TypesAndValues.TabularLibIpv6_Rfc0768Udp_TypesAndValues;
import com.testingtech.generated.codec.LibIpv6_Rfc2461NeighborDiscovery_TypesAndValues.TabularLibIpv6_Rfc2461NeighborDiscovery_TypesAndValues;
import com.testingtech.generated.codec.LibIpv6_Rfc2463Icmpv6_TypesAndValues.TabularLibIpv6_Rfc2463Icmpv6_TypesAndValues;
import com.testingtech.generated.codec.LibIpv6_Rfc2894RouterRenumbering_TypesAndValues.TabularLibIpv6_Rfc2894RouterRenumbering_TypesAndValues;
import com.testingtech.generated.codec.LibIpv6_Rfc3775Mipv6_TypesAndValues.TabularLibIpv6_Rfc3775Mipv6_TypesAndValues;
import com.testingtech.generated.codec.LibIpv6_Rfc792Icmpv4_TypesAndValues.TabularLibIpv6_Rfc792Icmpv4_TypesAndValues;
import com.testingtech.generated.codec.LibIpv6_Rfc826Arp_TypesAndValues.TabularLibIpv6_Rfc826Arp_TypesAndValues;
import com.testingtech.ttcn.logging.LoggingInterface;
import com.testingtech.ttcn.tci.TciEncoding;
import com.testingtech.ttcn.tci.TciModuleParameterIdImpl;
import com.testingtech.ttcn.tri.enums.IEnumDefinitions;
import com.testingtech.ttcn.tri.extFunc.ExternalFunctions_IPv6;
import com.testingtech.util.HexViewer;

import de.tu_berlin.cs.uebb.muttcn.runtime.RB;
import de.tu_berlin.cs.uebb.ttcn.runtime.OctetString;


/**
 * IPv6TestAdapter.java This is the test adapter for the IPv6 TS on
 * linux/windows platform(STF276).<br>
 * 
 * The configurations of the test adapter:
 * 
 * <ul>
 * <li>Receive hashmap:
 * <pre>
 * +-----------------------------------+-----------------------------------+
 * | Key                               | Entry                             |
 * +-----------------------------------+-----------------+-----------------+
 * | MAC address                       | ComponentId     | Role            |
 * +-----------------------------------+-----------------+-----------------+
 * </pre>
 * With the receive hashmap the test adapter gets all the information to 
 * retrieve the component to enqueue the message in accessing this 
 * information via the MAC address of the received message. In addition with 
 * the port hashmap the tsiPortId, needed to enqueue, can be retrieved with 
 * this componentId.<br><br>
 * 
 * <li>Send hashmap:
 * <pre>
 * +-----------------------------------+-----------------------------------+
 * | Key                               | Entry                             |
 * +-----------------+-----------------+-----------------+-----------------+
 * | IP address      | ComponentId     | NetDevice       | MAC address     |
 * +-----------------+-----------------+-----------------+-----------------+
 * </pre>
 * With the send hashmap the test adapter gets all the information to send 
 * the retrieved message to a configured net device. The information can be 
 * retrieved via the sent IP address.<br><br>
 * 
 * <li>port hashmap:
 * <pre>
 * +-----------------------------------+-----------------------------------+
 * | Key                               | Entry                             |
 * +-----------------------------------+-----------------------------------+
 * | CompIdPort                        | tsiPortId                         |
 * +-----------------------------------+-----------------------------------+
 * </pre>
 * With the port hashmap the test adapter gets all the information needed 
 * for enqueuing received messages.<br><br>
 * </ul>
 * 
 * @author <a href="mailto:tepelmann@testingtech.de">Dirk Tepelmann</a>
 * @version 0.1
 */
public class IPv6TestAdapter extends TestAdapter implements TriCommunicationSA,
TriPlatformPA, TciEncoding, JpcapHandler, TestAdapterDefinitions, IEnumDefinitions {
    private TciTMProvided management;
    private LoggingInterface logging = null;
    private RB RB = null;
    
    private static final String IPV6CODING = "LibIpv6_Interface_TypesAndValues";
    
    enum IpVersion {IPv4, IPv6};

    /**
     * HashMap declarations for the different information needed
     */
    /**
     * To store jpcap instance per net device, accessed via net device
     */
    private HashMap<String, Jpcap> jpcapMap = new HashMap<String, Jpcap>();
    /**
     * To store jpcap instance without filter per net device, accessed via net device
     */
    private HashMap<String, Jpcap> jpcapNoFilterMap = new HashMap<String, Jpcap>();
    /**
     * To store jpcapSender instance per net device, accessed via net device
     */
    private HashMap<String, JpcapSender> jpcapSenderMap = new HashMap<String, JpcapSender>();
    
    /**
     * To store net device id and MAC information, accessed via IP address and PTC ref
     */
    private HashMap<RoutingEntry, NetDeviceAndMAC> sendMap = new HashMap<RoutingEntry, NetDeviceAndMAC>();
    /**
     * To store PTC ref and role information, accessed via MAC address
     */
    private HashMap<BytesHash, PTCAndRole> rcvMap = new HashMap<BytesHash, PTCAndRole>();
    
    /**
     * To store filter information per net devie, accessed via net device  
     */
    private HashMap<String, String> filterMap = new HashMap<String, String>();
    
    /**
     * To store tsiPortId information, accessed via compPortId(ipPort)
     */
    private HashMap<TriPortIdImpl, TriPortId> portMap = new HashMap<TriPortIdImpl, TriPortId>();
    
    /**
     * To store jpcapRunnable instance, accessed via net device  
     */
    private HashMap<String, JpcapRunnable> jpcapRunnableMap = new HashMap<String, JpcapRunnable>();
    
    /**
     * To store jpcapWriterRunnable instance, accessed via net device  
     */
    private HashMap<String, JpcapWriterRunnable> jpcapWriterRunnableMap = new HashMap<String, JpcapWriterRunnable>();
    
    /**
     * To store security association information for ESP, accessed via security parameter index  
     */
    private HashMap<BigInteger, SecurityParameterInfo> saMap = new HashMap<BigInteger, SecurityParameterInfo>();
    
    /**
     * To store security association information for IKE.
     */
    private IKESecurityParameterInfo ikeSP = null;
    
    /**
     * The TCP socket for the upper tester communication
     */
    Socket scopSocket = null;
    
    private Object threadlock = new Object();
    private boolean runThread = false;
    private Thread scopReceiverThread = null;
    
    private final int MSG_BUF_SIZE = 1600;
    
    private final int DST = 0x0;
    private final int SRC = DST + 1;
    
    //TODO
	SimpleDateFormat df = new SimpleDateFormat("HH:mm:ss:SSS");
	Calendar cal = Calendar.getInstance();
	private boolean dbg = false;
	
	private String tcName = "";
	private boolean dump = false;
	private String dumpPath = "";
	
	/**
     * Creates a new IPv6TestAdapter object.
     */
    public IPv6TestAdapter() {
        super();
    }
    
    /* (non-Javadoc)
     * @see com.testingtech.ttcn.tri.TestAdapter#setRB(de.tu_berlin.cs.uebb.muttcn.runtime.RB)
     */
    public TestAdapter setRB(RB rb) {
        super.setRB(rb);
        RB = rb;
        management = RB.getTciTMProvided();
        logging = RB.getLogging();
        
        addExternalFunctionImpl(new ExternalFunctions_IPv6(RB));
        
        dbg = RB.debug;
//        dbg = true;
        
        return this;
    }
    
    /* (non-Javadoc)
     * @see com.testingtech.ttcn.tci.TciEncoding#getCodec(java.lang.String)
     */
    public TciCDProvided getCodec(String encodingName) {
        // RB check if decoding parameters in ExtFunction
        if ((encodingName == null) || encodingName.equals("")) {
            encodingName = IPV6CODING;
        }
        
        TciCDProvided codec = (TciCDProvided) super.getCodec(encodingName);
        
        if (codec != null) {
            return codec;
        }
        
        if (encodingName.equals(IPV6CODING)) { // if with encoding is not set.
            codec = new TabularLibIpv6_Interface_TypesAndValues(RB);
            codecs.put(IPV6CODING, codec);
        }
        else if (encodingName.equals("LibIpv6_Interface_TypesAndValues")) {
            codec = new TabularLibIpv6_Interface_TypesAndValues(RB);
            codecs.put(IPV6CODING, codec);
        }
//        else if (encodingName.equals("LibIpv6_Rfc2460Root_TypesAndValues")) {
//            codec = (TciCDProvided) codecs.get("LibIpv6_Rfc2460Root_TypesAndValues");
//            
//            if (codec != null) {
//                return codec;
//            }
//            
//            codec = new TabularLibIpv6_Rfc2460Root_TypesAndValues(RB);
//            codecs.put("LibIpv6_Rfc2460Root_TypesAndValues", codec);
//            
//            return codec;
//        }
        else if (encodingName.equals("LibIpv6_Rfc2461NeighborDiscovery_TypesAndValues")) {
            codec = (TciCDProvided) codecs.get("LibIpv6_Rfc2461NeighborDiscovery_TypesAndValues");
            
            if (codec != null) {
                return codec;
            }
            
            codec = new TabularLibIpv6_Rfc2461NeighborDiscovery_TypesAndValues(RB);
            codecs.put("LibIpv6_Rfc2461NeighborDiscovery_TypesAndValues", codec);
            
            return codec;
        }
        else if (encodingName.equals("LibIpv6_Rfc2463Icmpv6_TypesAndValues")) {
            codec = (TciCDProvided) codecs.get("LibIpv6_Rfc2463Icmpv6_TypesAndValues");
            
            if (codec != null) {
                return codec;
            }
            
            codec = new TabularLibIpv6_Rfc2463Icmpv6_TypesAndValues(RB);
            codecs.put("LibIpv6_Rfc2463Icmpv6_TypesAndValues", codec);
            
            return codec;
        }
        else if (encodingName.equals("LibIpv6_Rfc2894RouterRenumbering_TypesAndValues")) {
            codec = (TciCDProvided) codecs.get("LibIpv6_Rfc2894RouterRenumbering_TypesAndValues");
            
            if (codec != null) {
                return codec;
            }
            
            codec = new TabularLibIpv6_Rfc2894RouterRenumbering_TypesAndValues(RB);
            codecs.put("LibIpv6_Rfc2894RouterRenumbering_TypesAndValues", codec);
            
            return codec;
        }
        else if (encodingName.equals("LibIpv6_Rfc3775MIPv6_TypesAndValues")) {
            codec = (TciCDProvided) codecs.get("LibIpv6_Rfc3775MIPv6_TypesAndValues");
            
            if (codec != null) {
                return codec;
            }

            codec = new TabularLibIpv6_Rfc3775Mipv6_TypesAndValues(RB);
            codecs.put("LibIpv6_Rfc3775MIPv6_TypesAndValues", codec);
            
            return codec;
        }
//        else if (encodingName.equals("LibIpv6_MultiRfcs_TypesAndValues")) {
//            codec = (TciCDProvided) codecs.get("LibIpv6_MultiRfcs_TypesAndValues");
//            
//            if (codec != null) {
//                return codec;
//            }
//
//            codec = new TabularLibIpv6_MultiRfcs_TypesAndValues(RB);
//            codecs.put("LibIpv6_MultiRfcs_TypesAndValues", codec);
//            
//            return codec;
//        }
        else if (encodingName.equals("LibIpv6_CommonRfcs_TypesAndValues")) {
            codec = (TciCDProvided) codecs.get("LibIpv6_CommonRfcs_TypesAndValues");
            
            if (codec != null) {
                return codec;
            }

            codec = new TabularLibIpv6_CommonRfcs_TypesAndValues(RB);
            codecs.put("LibIpv6_CommonRfcs_TypesAndValues", codec);
            
            return codec;
        }
        else if (encodingName.equals("LibIpv6_Rfc0768Udp_TypesAndValues")) {
            codec = (TciCDProvided) codecs.get("LibIpv6_Rfc0768Udp_TypesAndValues");
            
            if (codec != null) {
                return codec;
            }
            
            codec = new TabularLibIpv6_Rfc0768Udp_TypesAndValues(RB);
            codecs.put("LibIpv6_Rfc0768Udp_TypesAndValues", codec);
            
            return codec;
        }
        else if (encodingName.equals("LibIpv6_Rfc792Icmpv4_TypesAndValues")) {
            codec = (TciCDProvided) codecs.get("LibIpv6_Rfc792Icmpv4_TypesAndValues");
            
            if (codec != null) {
                return codec;
            }
            
            codec = new TabularLibIpv6_Rfc792Icmpv4_TypesAndValues(RB);
            codecs.put("LibIpv6_Rfc792Icmpv4_TypesAndValues", codec);
            
            return codec;
        }
        else if (encodingName.equals("LibIpv6_Rfc826Arp_TypesAndValues")) {
            codec = (TciCDProvided) codecs.get("LibIpv6_Rfc826Arp_TypesAndValues");
            
            if (codec != null) {
                return codec;
            }
            
            codec = new TabularLibIpv6_Rfc826Arp_TypesAndValues(RB);
            codecs.put("LibIpv6_Rfc826Arp_TypesAndValues", codec);
            
            return codec;
        }
        else if (encodingName.equals("ExtFuncCodec")) {
            codec = super.getCodec("ExtFuncCodec");
            codecs.put("ExtFuncCodec", codec);
        }
        else {
            tciTMProvided.tciError("Unknown codec " + encodingName);
        }
        
        return codec;
    }
    
    /* (non-Javadoc)
     * @see org.etsi.ttcn.tri.TriCommunicationSA#triExecuteTestcase(org.etsi.ttcn.tri.TriTestCaseId, org.etsi.ttcn.tri.TriPortIdList)
     */
    public TriStatus triExecuteTestcase(TriTestCaseId testcase,
            TriPortIdList tsiList) {
        
    	tcName = testcase.getTestCaseName();
    	
    	try {
    		dump = ((BooleanValue)management.tciGetModulePar(TciModuleParameterIdImpl.valueOf("PX_WRITE_DUMP"))).getBoolean();
    	}
    	catch (Exception e) {
    		//modulepar not declared, so false anyway 
			dump = false;
		}
    	
    	try {
    		dumpPath = ((CharstringValue)management.tciGetModulePar(TciModuleParameterIdImpl.valueOf("PX_WRITE_DUMP_PATH"))).getString();
    	}
    	catch (Exception e) {
    		//modulepar not declared, so empty string anyway 
			dumpPath = "";
		}
    	
        return new TriStatusImpl();
    }  
    
    /* (non-Javadoc)
     * @see org.etsi.ttcn.tri.TriCommunicationSA#triMap(org.etsi.ttcn.tri.TriPortId, org.etsi.ttcn.tri.TriPortId)
     */
    public synchronized TriStatus triMap(final TriPortId compPortId, final TriPortId tsiPortId) {
        if (compPortId.getPortName().equals(IPPORT) 
        		|| compPortId.getPortName().equals(IP4PORT)
        		 || compPortId.getPortName().equals(ARPPORT)) {
        	TriPortIdImpl hashPortId = new TriPortIdImpl(compPortId.getPortName(), compPortId.getComponent(), compPortId.getPortIndex());
        	synchronized (portMap) {
            	portMap.put(hashPortId, tsiPortId);
			}
        }
        else if (compPortId.getPortName().equals(COMPSCOPPORT)){
        	if (scopSocket != null) {
        		return new TriStatusImpl("scopPort already mapped!");
        	}
        	else {
        		String host = "172.30.8.4";//((CharstringValue)RB.tciTMProvided.tciGetModulePar(TciModuleParameterIdImpl.valueOf("PX_SCOP_HOSTNAME"))).getString();
        		int port = 3751;//((IntegerValue)RB.tciTMProvided.tciGetModulePar(TciModuleParameterIdImpl.valueOf("PX_SCOP_PORT"))).getInt();
        		try {
					scopSocket = new Socket(host, port);
				} catch (UnknownHostException e) {
					return new TriStatusImpl(e.getMessage());
				} catch (IOException e) {
					return new TriStatusImpl(e.getMessage());
				}
				if (!scopSocket.isConnected()) {
					return new TriStatusImpl("Problems connecting to SCOP server");
				}
				
				/* Starting receiver thread */
				runThread = true;
				scopReceiverThread = new Thread(Thread.currentThread().getThreadGroup().getParent().getParent(), "SCOP receiver thread") {
					public void run() {
						boolean mylock = runThread;
						
						try {
							scopSocket.setSoTimeout(100);
							
							while (mylock) {
								byte [] msg = new byte[MSG_BUF_SIZE];
								
								try {
									int length = scopSocket.getInputStream().read(msg);
									
									if (length > 0) {
										byte [] receivedMessage = new byte[length];
							            System.arraycopy(msg, 0, receivedMessage, 0, length);
							            if (dbg)
							            	System.err.println("Received SCOP message: " + HexViewer.hexString(receivedMessage));
										TriMessage rcvMessage = new TriMessageImpl(receivedMessage);
										
										synchronized (threadlock) {
											if (runThread) {
												enqueueMsg(tsiPortId, null, compPortId.getComponent(), rcvMessage);
											}
										}
									}
								}
								catch (InterruptedIOException ioe) {
									//
								}
								catch (IOException ioe) {
									log(ioe.getMessage());
									if (scopSocket != null) {
										scopSocket.close();
										scopSocket = null;
									}
									
									return;
								}
								
								synchronized (threadlock) {
									mylock = runThread;
								}
							}
							
							if (scopSocket != null) {
								scopSocket.close();
								scopSocket = null;
							}
						}
						catch (IOException ioe) {
							log(ioe.getMessage());
						}
					}
				};
				scopReceiverThread.start();
        	}
        }
        if (!(tsiPortId.getPortName().equals(CFG_TA_PORT)
				|| tsiPortId.getPortName().equals(ETHERNETPORT) 
				|| tsiPortId.getPortName().equals(IPV4ETHERNETPORT) 
				|| tsiPortId.getPortName().equals(ARPETHERNETPORT) 
				|| tsiPortId.getPortName().equals(TSISCOPPORT))
        ) {
			return new TriStatusImpl(tsiPortId.getPortName() + " not supported by this TA. " +
					"Only " + CFG_TA_PORT + ", " + ETHERNETPORT  + ", " + IPV4ETHERNETPORT  + ", " + ARPETHERNETPORT + " and " + TSISCOPPORT + " are supported.");
        }
        
        return super.triMap(compPortId, tsiPortId);
    }
    /* (non-Javadoc)
     * @see org.etsi.ttcn.tri.TriCommunicationSA#triUnmap(org.etsi.ttcn.tri.TriPortId, org.etsi.ttcn.tri.TriPortId)
     */
    public synchronized TriStatus triUnmap(TriPortId compPortId, TriPortId tsiPortId) {
    	TriStatus status = super.triUnmap(compPortId, tsiPortId);
    	
        if (compPortId.getPortName().equals(IPPORT)
				|| compPortId.getPortName().equals(IP4PORT)
				|| compPortId.getPortName().equals(ARPPORT)) {
        	synchronized (portMap) {
            	TriPortIdImpl hashPortId = new TriPortIdImpl(compPortId.getPortName(), compPortId.getComponent(), compPortId.getPortIndex());
                portMap.remove(hashPortId);
			}
        }
        else if (compPortId.getPortName().equals(COMPSCOPPORT)) {
        	if (scopSocket.isConnected()) {
        		status = scopCancel();
        	}
        }
        
        return status;
    }
    
    /**
     * Enqueues the message to the TE.
     * 
     * @param sender
     * @param address
     * @param receiver
     * @param message
     */
    public synchronized void enqueueMsg(TriPortId sender, TriAddress address,
            TriComponentId receiver, TriMessage message) {
        Cte.triEnqueueMsg(sender, address, receiver, message);
    }
    
    /**
     * Sets the filter for capturing messages from the net device and
     * starts the capturing.
     * 
     * @param jpcap 		The instance of capturing
     * @param netDeviceId	The net device for capturing
     * @param macFilterList	The MAC address list for filtering
     * @param ipVersion TODO
     */
    public synchronized void setFilterAndStartThread(Jpcap jpcap, String netDeviceId, RecordOfValue macFilterList, Enum ipVersion) {
        Thread th = null;
        
        JpcapRunnable jr;
        synchronized (jpcapRunnableMap) {
			jr = (JpcapRunnable)jpcapRunnableMap.get(netDeviceId);
		}
        if (jr == null) {
            jr = new JpcapRunnable(RB, jpcap, netDeviceId, portMap, rcvMap);
            synchronized (jpcapRunnableMap) {
            	jpcapRunnableMap.put(netDeviceId, jr);
            }
        }
        else {
        	if (dbg) {
        		cal.setTimeInMillis(System.currentTimeMillis());
        		System.err.println(df.format(cal.getTime()) + "-------------------------------cancel jpcapRunnable for " + netDeviceId + "------------------------------");
        	}
            synchronized (jr) {
                try {
                    jr.cancel();
					jr.wait();
		            //FIXME possible endless loop
//                while(jr.isRunning());
				} catch (InterruptedException e1) {
					return;
				}
			}
            if (dbg) {
            	cal.setTimeInMillis(System.currentTimeMillis());
            	System.err.println(df.format(cal.getTime()) + "-------------------------------cancel jpcapRunnable done? ------------------------------" + jr);
            }
            //update information about current hashmaps.
            jr.setPortMap(portMap);
            jr.setRcvHashMap(rcvMap);
        }

        String filter;
		StringBuffer strBuf = null;

		synchronized (filterMap) {
			filter = (String)filterMap.get(netDeviceId);
		}

		String ipFilterVersion = "";
		if (ipVersion == IpVersion.IPv6) {
			ipFilterVersion = "ip6";
		}
		else {
			ipFilterVersion = "arp or ip";
		}
		
		//enhance the filter
        if (filter == null) {
			//            filter = "ip6 and (";
            filter = ipFilterVersion + " and () and not(";
			strBuf = new StringBuffer(filter);
        }
        else {
			strBuf = new StringBuffer(filter);
			int lio = strBuf.lastIndexOf(") and not");
			strBuf.insert(lio, " or ");
			strBuf.deleteCharAt(strBuf.length()-1);
			strBuf.append(" or ");
        }
//        else {
//            StringBuffer strBuf = new StringBuffer(filter);
//            strBuf.deleteCharAt(strBuf.length()-1);
//            filter = strBuf.toString() + " or ";
//        }
        int len = macFilterList.getLength();
        
        try {
			for (int i=0; i<len; i++) {
				int lio = strBuf.lastIndexOf(") and not");
        		if (i<(len-1)) {
        			strBuf.insert(lio, getEtherFilter((OctetstringValue)macFilterList.getField(i), DST) + " or ");
        		}
				else {
        			strBuf.insert(lio, getEtherFilter((OctetstringValue)macFilterList.getField(i), DST));
        		}
			}
			for (int i=0; i<len; i++) {
        		if (i<(len-1)) {
        			strBuf.append(getEtherFilter((OctetstringValue)macFilterList.getField(i), SRC) + " or ");
        		}
				else {
        			strBuf.append(getEtherFilter((OctetstringValue)macFilterList.getField(i), SRC));
        		}
			}
			strBuf.append(")");
			filter = strBuf.toString();
//        try {
//        	for (int i=0; i<len; i++) {
//        		filter += getEtherFilter((OctetstringValue)macFilterList.getField(i), DST);
//        		if (i<(len-1)) {
//        			filter += " or ";
//        		}
//        	}
//        	filter += ") and not (";
//        	for (int i=0; i<len; i++) {
//        		filter += getEtherFilter((OctetstringValue)macFilterList.getField(i), SRC);
//        		if (i<(len-1)) {
//        			filter += " or ";
//        		}
//        	}
//        	filter += ")";
            //TODO
//            filter = "ip6";
			
        	jpcap.setFilter(filter, false);
        	
        	if (dbg) {
        		System.out.println("filter: " + filter);
        	}
        	
        	synchronized (filterMap) {
            	filterMap.put(netDeviceId, filter);        
			}
        	
        	th = new Thread(Thread.currentThread().getThreadGroup().getParent().getParent(), jr, "Receiver thread");
        	th.start();
        	//TODO
        	if (dbg) {
        		cal.setTimeInMillis(System.currentTimeMillis());
        		System.err.println(df.format(cal.getTime()) + "-------------------------------jpcapRunnableMap start for " + netDeviceId + "------------------------------" + jr);
        	}
        }
        catch (Exception e) {
        	management.tciError("Filter not set: " + e.getMessage());
        }
    }
    
    /* (non-Javadoc)
     * @see org.etsi.ttcn.tri.TriCommunicationSA#triSend(org.etsi.ttcn.tri.TriComponentId, org.etsi.ttcn.tri.TriPortId, org.etsi.ttcn.tri.TriAddress, org.etsi.ttcn.tri.TriMessage)
     */
    public synchronized TriStatus triSend(TriComponentId componentId, TriPortId tsiPortId,
            TriAddress address, TriMessage message) {
        
    	TriStatus rstatus = new TriStatusImpl();
    	
        if (tsiPortId.getPortName().equals(CFG_TA_PORT)) {
        	/* Prepare the configurations based on the information included in the config messages */
        	
            Value value = getCodec("ExtFuncCodec").decode(message, null);
            
            String typeName = value.getType().getName();
            
            if (typeName.equals(CfgMessage)){
            	String role = "";
                String netDeviceId = "";
                RecordOfValue macFilterList = null;
                RecordValue ipNodeParam = null;
                PTCAndRole par = null;
                
                IpVersion ipVersion = IpVersion.IPv6;

                /* Prepare the rcvMap and the sendMap */
                if (!((RecordValue)value).getField(ROLE).notPresent()) {
                    role = ((CharstringValue)((RecordValue)value).getField(ROLE)).getString();
                    par = new PTCAndRole(componentId, role);
                }
                else {
                    return new TriStatusImpl(typeName + " does not contain mandatory " + ROLE + " field");
                }
                
                //Check if we run IPv4 or IPv6
                if (role.startsWith(IPV4ROUTER_ROLE)) {
                	ipVersion = IpVersion.IPv4;
                }
                
                if (!((RecordValue)value).getField(NETDEVICEID).notPresent()) {
                    netDeviceId = ((CharstringValue)((RecordValue)value).getField(NETDEVICEID)).getString();
                }
                else {
                    return new TriStatusImpl(value.getType().getName() + " does not contain mandatory " + NETDEVICEID + " field");
                }

                for (int i=0; i<IPv6NODEPARAMSLIST.length; i++) {
                	ipNodeParam = (RecordValue)((RecordValue)value).getField(IPv6NODEPARAMSLIST[i]);
                	if (dbg)
                		System.out.println(IPv6NODEPARAMSLIST[i]);
                	if (!ipNodeParam.notPresent()) {
                		if (dbg)
                			System.out.println(IPv6NODEPARAMSLIST[i] + "is present");
                        if (!(ipNodeParam).getField(HANDLECFG).notPresent() 
                        		&& (((BooleanValue)(ipNodeParam).getField(HANDLECFG)).getBoolean()) == true) {
                        	rstatus = handleConfig(ipNodeParam, netDeviceId, par, IPv6NODEPARAMSLIST[i].equals("paramsIut"), IPv6NODEPARAMS);
                        	if (rstatus.getStatus() == TriStatus.TRI_ERROR) {
                        		return rstatus;
                        	}
                        }
                	}
                }
                
                /* Retrieve the MAC filter list */
                if (!((RecordValue)value).getField(MACFILTERLIST).notPresent()) {
                    macFilterList = (RecordOfValue)((RecordValue)value).getField(MACFILTERLIST);
                }
                else {
                    return new TriStatusImpl(typeName + " does not contain mandatory " + MACFILTERLIST + " field");
                }
                
                /* Configure Jpcap */
                return handleJpcapWithFilter(netDeviceId, macFilterList, ipVersion);
                
            }
            else {
                return new TriStatusImpl("unknown type at " + tsiPortId.getPortName());
            }
        }
        else if (tsiPortId.getPortName().equals(ETHERNETPORT) || tsiPortId.getPortName().equals(IPV4ETHERNETPORT)){
        	/* Send messages via Ethernet */
        	
            byte[] msg = message.getEncodedMessage();

            /* Create a Ethernet message based of the configured information
             * and the message sent by the TE
             */
            EthernetPacket ethP = new EthernetPacket();
            
            byte [] srcIpAddress = getSrcIPAddress(msg);

            RoutingEntry re = new RoutingEntry(srcIpAddress, componentId);
            NetDeviceAndMAC ndam;
            synchronized (sendMap) {
				ndam = (NetDeviceAndMAC)sendMap.get(re);
			}
            if (ndam == null) {
                return new TriStatusImpl(TRISEND_ND_MAC_CONFIGURE_ERROR + ": " + new OctetString(RB, srcIpAddress.length, srcIpAddress).getString());
            }
            String netDeviceId = ndam.getNetDeviceId();
            if ((netDeviceId == null) || (netDeviceId.equals(""))) {
                return new TriStatusImpl(TRISEND_ND_CONFIGURE_ERROR + ": " + netDeviceId);
            }
            
            if (Arrays.equals(srcIpAddress, IP6_ALLNODESMCA)) {
            	ethP.src_mac = MAC_ALLNODESMCA;
            }
            else if (Arrays.equals(srcIpAddress, IP6_ALLROUTERSMCA)){
            	ethP.src_mac = MAC_ALLROUTERSMCA;
            }
            else {
            	ethP.src_mac = ndam.getMAC();
            	if (ethP.src_mac == null) {
            		return new TriStatusImpl("IPAddress(" + getIPv6AddressString(srcIpAddress) + "): " + TRISEND_SRC_IPADDRESS_CONFIGURE_ERROR);
            	}
            }
        	byte [] dstIpAddress = getDstIPAddress(msg);
//        	System.out.println(DebugHelp.hexString(dstIpAddress));
        	if (Arrays.equals(dstIpAddress, IP6_ALLNODESMCA)) {
//        		System.out.println("is IP_ALLNODESMCA");
        		ethP.dst_mac = MAC_ALLNODESMCA;
        	}
        	else if (Arrays.equals(dstIpAddress, IP6_ALLROUTERSMCA)){
//        		System.out.println("is IP_ALLROUTERSMCA");
        		ethP.dst_mac = MAC_ALLROUTERSMCA;
        	}
        	else {
        		RoutingEntry re_dst = new RoutingEntry(dstIpAddress, componentId);
        		NetDeviceAndMAC ndam_dst;
        		synchronized (sendMap) {
					ndam_dst = (NetDeviceAndMAC)sendMap.get(re_dst);
				}
        		if (ndam_dst == null) {
        			//TODO
        			return new TriStatusImpl("IPAddress(" + getIPv6AddressString(dstIpAddress) + "): " + TRISEND_DST_IPADDRESS_CONFIGURE_ERROR);
        		}
        		ethP.dst_mac = ndam_dst.getMAC();
        		if (ethP.dst_mac == null) {
        			return new TriStatusImpl("IPAddress(" + getIPv6AddressString(dstIpAddress) + "): " + TRISEND_DST_IPADDRESS_CONFIGURE_ERROR);
        		}
            }
            
//        	System.out.println("\nComponent(" + componentId.getComponentId() + "): \nSending message to netDeviceId " + netDeviceId +
//        			", ipaddress " + HexViewer.hexString(dstIpAddress) + "(src " + HexViewer.hexString(srcIpAddress) + 
//					") \nand MAC address " + HexViewer.hexString(ethP.dst_mac) + "(src " + HexViewer.hexString(ethP.src_mac) + ")");
        	
        	if ((msg[0]&0xF0) == 0x60) {
                ethP.frametype = IPv6FRAMETYPE;
        	}
        	else {
                ethP.frametype = IPv4FRAMETYPE;
        	}
        	
            ethP.data = new byte[msg.length];
            System.arraycopy(msg, 0, ethP.data, 0, msg.length);

            JpcapSender jpcapSender;
            synchronized (jpcapSenderMap) {
				jpcapSender = (JpcapSender)jpcapSenderMap.get(netDeviceId);
			}
            
//            System.out.println("jpcapSender: " + jpcapSender);
            
            /* Send the Ethernet message */
            if (jpcapSender != null) {
                jpcapSender.sendEthPacket(ethP);
            }
            else {
                return new TriStatusImpl(TRISEND_JPCAPSENDER_ERROR);
            }
        }
        else if (tsiPortId.getPortName().equals(ARPETHERNETPORT)){
        	/* Send messages via Ethernet */
        	
            byte[] msg = message.getEncodedMessage();

            /* Check for hardware type equals Ethernet */
            if (!((msg[0]==0x00) && (msg[1]==0x01))) {
            	return new TriStatusImpl("Only ARP via Ethernet supported! Do not send message!");
            }
            
            /* Create a Ethernet message based of the configured information
             * and the message sent by the TE
             */
            EthernetPacket ethP = new EthernetPacket();
            
            int hwAddrLength = msg[4];
            int protAddrLength = msg[5];

            byte [] srcIpAddress = new byte[protAddrLength];
            System.arraycopy(msg, 8 + hwAddrLength, srcIpAddress, 0, protAddrLength);

            RoutingEntry re = new RoutingEntry(srcIpAddress, componentId);
            NetDeviceAndMAC ndam;
            synchronized (sendMap) {
				ndam = (NetDeviceAndMAC)sendMap.get(re);
			}
            if (ndam == null) {
                return new TriStatusImpl(TRISEND_ND_MAC_CONFIGURE_ERROR + ": " + new OctetString(RB, srcIpAddress.length, srcIpAddress).getString());
            }
            String netDeviceId = ndam.getNetDeviceId();
            if ((netDeviceId == null) || (netDeviceId.equals(""))) {
                return new TriStatusImpl(TRISEND_ND_CONFIGURE_ERROR + ": " + netDeviceId);
            }

            ethP.src_mac = new byte[hwAddrLength];
            ethP.dst_mac = new byte[hwAddrLength];
            System.arraycopy(msg, 8, ethP.src_mac, 0, hwAddrLength);
            System.arraycopy(msg, 8 + hwAddrLength + protAddrLength, ethP.dst_mac, 0, hwAddrLength);
            
        	ethP.frametype = ARPFRAMETYPE;
        	
            ethP.data = new byte[msg.length];
            System.arraycopy(msg, 0, ethP.data, 0, msg.length);

            JpcapSender jpcapSender;
            synchronized (jpcapSenderMap) {
				jpcapSender = (JpcapSender)jpcapSenderMap.get(netDeviceId);
			}
            
            /* Send the Ethernet message */
            if (jpcapSender != null) {
                jpcapSender.sendEthPacket(ethP);
            }
            else {
                return new TriStatusImpl(TRISEND_JPCAPSENDER_ERROR);
            }
        }
        else if (tsiPortId.getPortName().equals(TSISCOPPORT)) {
        	/* Send upper layer messages */
        	try {
	        	byte [] msg = message.getEncodedMessage();
	        	
	        	scopSocket.getOutputStream().write(msg);
        	}
        	catch (IOException ioe) {
        		return new TriStatusImpl(ioe.getMessage());
        	}
        }
        else {
        	return new TriStatusImpl("TestAdapter does not support the port " + tsiPortId.getPortName());
        }
        
        return new TriStatusImpl();
    }

	/**
	 * @param netDeviceId
	 * @param macFilterList
	 * @param ipVersion TODO
	 */
	private TriStatus handleJpcapWithFilter(String netDeviceId, RecordOfValue macFilterList, Enum ipVersion) {
		Jpcap jpcap;
		synchronized (jpcapMap) {
			jpcap = (Jpcap)jpcapMap.get(netDeviceId);
		}
		Jpcap jpcapNoF = null;
		if (dump) {
			synchronized (jpcapNoFilterMap) {
				jpcapNoF = (Jpcap)jpcapNoFilterMap.get(netDeviceId);
			}
		}
		JpcapSender jpcapSender;
		synchronized (jpcapSenderMap) {
			jpcapSender = (JpcapSender)jpcapSenderMap.get(netDeviceId); 
		}
		try {
		    if (jpcap == null) {
		    	if (dbg)
		    		System.err.println("@@@@@@@@@@@@@@@@@ Creating new Jpcap @@@@@@@@@@@@@@@@@");
		        jpcap = Jpcap.openDevice(netDeviceId, 64000, true, -1);
		        jpcap.setNonBlocking(true);
		        synchronized (jpcapMap) {
		            jpcapMap.put(netDeviceId, jpcap);
				}
		    }
		    if (dump) {
		        if (jpcapNoF == null) {
		        	if (dbg)
		        		System.err.println("@@@@@@@@@@@@@@@@@ Creating new JpcapNoF @@@@@@@@@@@@@@@@@");
		            jpcapNoF = Jpcap.openDevice(netDeviceId, 64000, true, -1);
		            jpcapNoF.setNonBlocking(true);
		            synchronized (jpcapNoFilterMap) {
		                jpcapNoFilterMap.put(netDeviceId, jpcapNoF);
					}
		        }
		        JpcapWriterRunnable jwr = null;
		        synchronized (jpcapWriterRunnableMap) {
					jwr = (JpcapWriterRunnable)jpcapWriterRunnableMap.get(netDeviceId);
				}
		        //only initialised the first time
		        if (jwr == null) {
		            jwr = new JpcapWriterRunnable(jpcapNoF, dumpPath, tcName + "_" + netDeviceId + ".cap");
		            synchronized (jpcapWriterRunnableMap) {
		            	jpcapWriterRunnableMap.put(netDeviceId, jwr);
		            }
		            Thread th = new Thread(Thread.currentThread().getThreadGroup().getParent().getParent(), jwr, "Write capture thread");
		        	th.start();
		        }
		    }
		    if (jpcapSender == null) {
		        jpcapSender = JpcapSender.openDevice(netDeviceId);
		        synchronized (jpcapSenderMap) {
		            jpcapSenderMap.put(netDeviceId, jpcapSender);
				}
//                    	System.out.println("Component(" + componentId.getComponentId() + "): OpenDevice(" + netDeviceId + "), jpcapSender(" + jpcapSender + ")");
		    }
		    
		    setFilterAndStartThread(jpcap, netDeviceId, macFilterList, ipVersion);
		    
		    return new TriStatusImpl();
		}
		catch (IOException ioex) {
		    RB.getTciTMProvided().tciError("IPv6TestAdapter: " + ioex.getMessage());
		    
		    return new TriStatusImpl(ioex.getMessage());
		}
	}
    
    /* (non-Javadoc)
     * @see org.etsi.ttcn.tri.TriCommunicationSA#triSutActionInformal(java.lang.String)
     */
// We will use default SUTaction with PopUp-dialog instead
//    public TriStatus triSutActionInformal(String description) {
//        log(description);
//        
//        return new TriStatusImpl();
//    }
    
    
	/* (non-Javadoc)
	 * @see org.etsi.ttcn.tri.TriCommunicationSA#triSutActionTemplate(org.etsi.ttcn.tri.TriActionTemplate)
	 */
	public TriStatus triSutActionTemplate(TriActionTemplate at) {
        log(at.toString());
        
        return new TriStatusImpl();
	}
	
    /* (non-Javadoc)
     * @see org.etsi.ttcn.tri.TriCommunicationSA#triSAReset()
     */
    public synchronized TriStatus triSAReset() {
    	synchronized (portMap) {
            portMap.clear();
		}

        clearConfiguration(true);
        
        if (scopSocket != null) {
        	if (scopSocket.isConnected()) {
        		scopCancel();
        	}
        }
        
        return super.triSAReset();
    }
    
    /* (non-Javadoc)
	 * @see com.testingtech.ttcn.tri.TestAdapter#triExternalFunction(org.etsi.ttcn.tri.TriFunctionId, org.etsi.ttcn.tri.TriParameterList, org.etsi.ttcn.tri.TriParameter)
	 */
	public synchronized TriStatus triExternalFunction(TriFunctionId functionId, TriParameterList parameterList, TriParameter returnValue) {
		return super.triExternalFunction(functionId, parameterList, returnValue);
	}

	/**
     * Clears the configuration.
     */
    private synchronized void clearConfiguration(boolean full) {
        /* Remove all established configurations and their information */
    	
    	synchronized (jpcapRunnableMap) {
            for (Iterator it=jpcapRunnableMap.entrySet().iterator(); it.hasNext(); ) {
                Map.Entry entry = (Map.Entry)it.next();
                JpcapRunnable jr = ((JpcapRunnable)entry.getValue());
                if (dbg) {
                	cal.setTimeInMillis(System.currentTimeMillis());
                	System.err.println(df.format(cal.getTime()) + "-------------------------------clearConfiguration jpcapRunnableMap cancel------------------------------");
                }
                synchronized (jr) {
                	try {
                        //FIXME possible endless loop
						//                    while(jr.isRunning());
                        jr.cancel();
						jr.wait();
					} catch (InterruptedException e) {
						return;
					}
				}
                if (dbg) {
                	cal.setTimeInMillis(System.currentTimeMillis());
                	System.err.println(df.format(cal.getTime()) + "-------------------------------clearConfiguration jpcapRunnableMap cancel done? ------------------------------");
                }
            }

            jpcapRunnableMap.clear();
		}
    	
    	if (full) {
	    	synchronized (rcvMap) {
	            rcvMap.clear();
			}
	    	synchronized (sendMap) {
	            sendMap.clear();
			}
	    	synchronized (filterMap) {
	            filterMap.clear();
			}

	    	synchronized (jpcapWriterRunnableMap) {
	            for (Iterator it=jpcapWriterRunnableMap.entrySet().iterator(); it.hasNext(); ) {
	                Map.Entry entry = (Map.Entry)it.next();
	                JpcapWriterRunnable jwr = ((JpcapWriterRunnable)entry.getValue());
	                synchronized (jwr) {
	                	try {
	                        jwr.cancel();
							jwr.wait();
						} catch (InterruptedException e) {
							return;
						}
					}
	            }

	            jpcapWriterRunnableMap.clear();
			}

	    	synchronized (jpcapNoFilterMap) {
	            for (Iterator it=jpcapNoFilterMap.entrySet().iterator(); it.hasNext(); ) {
	                Map.Entry entry = (Map.Entry)it.next();
	                ((Jpcap)entry.getValue()).myClose();
	            }
	            jpcapNoFilterMap.clear();
			}

    	}
    	
    	synchronized (jpcapMap) {
    		//FIXME inserted due to commented out close at end of thread
            for (Iterator it=jpcapMap.entrySet().iterator(); it.hasNext(); ) {
                Map.Entry entry = (Map.Entry)it.next();
                if (dbg)
                	System.out.println("clearConfiguration Closing jpcap");
                ((Jpcap)entry.getValue()).myClose();
            }
    		//Closing is already done at end of jpcapRunnable thread
            jpcapMap.clear();
		}

    	synchronized (jpcapSenderMap) {
            for (Iterator it=jpcapSenderMap.entrySet().iterator(); it.hasNext(); ) {
                Map.Entry entry = (Map.Entry)it.next();
                if (dbg)
                	System.out.println("Closing " + ((JpcapSender)entry.getValue()).getDeviceId());
                ((JpcapSender)entry.getValue()).close();
            }
            jpcapSenderMap.clear();
		}
    }

    /* (non-Javadoc)
     * @see jpcap.JpcapHandler#handlePacket(jpcap.Packet)
     */
    public void handlePacket(Packet packet) {
    	//nothing to do, just make Jpcap happy
    }
    
    /**
     * Logs TestAdapter events
     * 
     * @param logMessage
     */
    public synchronized void log(String logMessage) {
        System.err.println(logMessage);
        logging.logTestAdapter(new Date().getTime(), this, logMessage);
    }
    
    /**
     * Retrieves the byte array from a TCI OctetstringValue
     * 
     * @param len
     * @param ov
     * @return octetstring as byte array
     */
    private synchronized byte[] getOctetstring(OctetstringValue ov) {
        int octLen = ov.getLength();
        byte []myByte = new byte[octLen];
        for (int i=0; i<octLen; i++) {
            myByte[i] = (byte)(ov.getOctet(i));
        }
        
        return myByte;
    }
    
    /**
     * Retrieves the source IP address from a given packet
     * 
     * @param msg The packet
     * @return
     */
    private synchronized byte [] getSrcIPAddress(byte [] msg){
        byte []ipAddress = null;
        if ((msg[0]&0xF0) == 0x60) {
    		ipAddress = new byte[16];
            System.arraycopy(msg, 8, ipAddress, 0, 16);
        }
        else {
    		ipAddress = new byte[4];
            System.arraycopy(msg, 12, ipAddress, 0, 4);
        }
        
        return ipAddress;
    }
    
    /**
     * Retrieves the destination IP address from a given packet
     * 
     * @param msg
     * @return
     */
    private synchronized byte [] getDstIPAddress(byte [] msg){
        byte []ipAddress = null;
        if ((msg[0]&0xF0) == 0x60) {
        	ipAddress = new byte[16];
        	System.arraycopy(msg, 24, ipAddress, 0, 16);
        }
        else {
        	ipAddress = new byte[4];
        	System.arraycopy(msg, 16, ipAddress, 0, 4);
        }
        
        return ipAddress;
    }
     
    /**
     * Prepares a IP address string from a given IP packet
     * 
     * @param msg	The IP packet
     * @return		The IP address string	
     */
    public synchronized String getIPv6AddressString(byte [] msg){
    	String str = "";
    	
    	int len = msg.length;
    	for (int i=0; i<len; i++) {
    		//find out, how to set 
    		int tmp = new Integer(0xFF&msg[i]).intValue();
    		if (tmp == 0) {
    			str += Integer.toHexString(tmp);
    		}
    		else if (tmp<10) {
    			str += Integer.toHexString(0);
    		}
			str += Integer.toHexString(tmp);
    		if ((i%2 != 0) && (i!=(len -1))) {
    			str += ":";
    		}
    	}
    	
        return str;
    }
     
    /**
     * Prepares the filter string for capturing based on the TCI OctetstringValue.
     * 
     * @param ov
     * @return
     */
//    private synchronized String getEtherFilter(OctetstringValue ov) throws Exception {
    private synchronized String getEtherFilter(OctetstringValue ov, int dir) throws Exception {
        StringBuffer strBuf = new StringBuffer(((OctetString)ov).getStringValue());
        int len = strBuf.length();
        int cnt = 0;
        for (int i=2; i<len; i+=2) {
            strBuf.insert(i + cnt++, ':');
        }
        if (len!=12) {
        	throw new Exception("Invalid Ethernet-MACadress(" + strBuf + ") of length " + len/2 + " given(should be 6)");
        }
        //return "ether dst host " + strBuf.toString();
        if (dir == SRC) {
        	return "ether src host " + strBuf.toString();
        }
    	return "ether dst host " + strBuf.toString();
    }

    /**
     * Prepares the send and receive HashMaps for the given configurations.
     * 
     * @param ipv6Param		The IPv6 Parameter given in the config message
     * @param ipField		The name of the IP field
     * @param macField		The name of the MAC field
     * @param netDeviceId	The net device to store
     * @param par			The componentId included to enter into sendMap, the entry for the rcvMap
     * @return				Returns a TRI status.
     */
    public synchronized TriStatus prepareHashMaps(Value ipParam, String ipField, String macField, String netDeviceId, PTCAndRole par, boolean isIut) {
    	RoutingEntry re = null;

//    	if (!ipField.equals("haGlas")) {
	    	if (dbg) {
	    			System.err.println(ipField + ": IPaddress = " + HexViewer.hexString(getOctetstring(((OctetstringValue)((RecordValue)ipParam).getField(ipField)))));
		    		System.err.println("component: " + par.getComponentId().getComponentId());
	    	}
			re = new RoutingEntry(getOctetstring(((OctetstringValue)((RecordValue)ipParam).getField(ipField))), par.getComponentId());
//		}
//		else {
//			RecordOfValue rov = (RecordOfValue)((RecordValue)ipv6Param).getField(ipField);
//			for (int i=0; i<rov.getLength(); i++) {
//		    	if (dbg) {
//		    		System.err.println(ipField + ": IPaddress = " + HexViewer.hexString(getOctetstring(((OctetstringValue)((RecordOfValue)rov).getField(i)))));
//		    		System.err.println("component: " + par.getComponentId().getComponentId());
//		    	}
//		    	re = new RoutingEntry(getOctetstring(((OctetstringValue)((RecordOfValue)rov).getField(i))), par.getComponentId());
//			}
//		}

        BytesHash mac = new BytesHash(getOctetstring(((OctetstringValue)((RecordValue)ipParam).getField(macField))));
    	if (dbg) {
    		System.out.println(macField + ": macAddress = " + HexViewer.hexString(mac.getBytes()));
    	}
		NetDeviceAndMAC ndam = new NetDeviceAndMAC(netDeviceId, mac.getBytes());
		synchronized (sendMap) {
			sendMap.put(re, ndam);
			if (dbg)
				System.out.println("sendMap: " + sendMap);
		}
		if (!isIut) {
			synchronized (rcvMap) {
				rcvMap.put(mac, par);
				if (dbg)
					System.out.println("rcvMap: " + rcvMap);
			}
		}
		
        return new TriStatusImpl();
    }

    private synchronized TriStatus handleConfig(RecordValue ipParam, String netDeviceId, PTCAndRole par, boolean isIut, String[][] ipNodeParams) {
    	for (int i=0; i<ipNodeParams.length; i++) {
        	if (!(ipParam).getField(ipNodeParams[i][FADDR]).notPresent()) {
        		if (!(ipParam).getField(ipNodeParams[i][FADDRMAC]).notPresent()) {
        			TriStatus status = prepareHashMaps(ipParam, ipNodeParams[i][FADDR], ipNodeParams[i][FADDRMAC], netDeviceId, par, isIut);
                	if (status.getStatus() == TriStatus.TRI_ERROR) {
                		return status;
                	}
       		
        			if (dbg) {
        				System.out.println(ipNodeParams[i][FADDR] + " and "+ ipNodeParams[i][FADDRMAC] + " set");
        			}
        		}
        		else {
        			return new TriStatusImpl("Missing the " + ipNodeParams[i][FADDR] + " entry for " + ipNodeParams[i][FADDRMAC]);
	        	}
	        }
    	}
//    	for (int i=0; i<IPv6NODEPARAMS.length; i++) {
//        	if (!(ipv6Param).getField(IPv6NODEPARAMS[i][FADDR]).notPresent()) {
//        		if (!(ipv6Param).getField(IPv6NODEPARAMS[i][FADDRMAC]).notPresent()) {
//        			TriStatus status = prepareHashMaps(ipv6Param, IPv6NODEPARAMS[i][FADDR], IPv6NODEPARAMS[i][FADDRMAC], netDeviceId, par, isIut);
//                	if (status.getStatus() == TriStatus.TRI_ERROR) {
//                		return status;
//                	}
//       		
//        			if (dbg) {
//        				System.out.println(IPv6NODEPARAMS[i][FADDR] + " and "+ IPv6NODEPARAMS[i][FADDRMAC] + " set");
//        			}
//        		}
//        		else {
//        			return new TriStatusImpl("Missing the " + IPv6NODEPARAMS[i][FADDR] + " entry for " + IPv6NODEPARAMS[i][FADDRMAC]);
//	        	}
//	        }
//    	}
    	
        return new TriStatusImpl();
    }
    
    /**
     * Stops the receiver thread for the upper layer.
     * 
     * @return The TriStatus
     */
    public synchronized TriStatus scopCancel() {
    	synchronized (threadlock) {
    		if (!runThread) {
    			return new TriStatusImpl();
    		}
    		
    		runThread = false;
    	}
    	
    	while (scopReceiverThread.isAlive()) {
    		try {
    			Thread.sleep(500);
    		}
    		catch (InterruptedException ie) {
    			//
    		}
    	}
    	
    	if (scopSocket != null) {
    		try {
            	if (dbg) {
            		System.err.println("Closing socket");
            	}
    			scopSocket.close();
    		}
    		catch (IOException ioe) {
    			return new TriStatusImpl(ioe.getMessage());
    		}
    	}
    	
    	scopSocket = null;

        return new TriStatusImpl();
    }

    /**
     * @param spi
     * @param seqNbr
     * @param ahIntegrityAlgo
     * @param ahIntegrityKey
     * @param ahIcvLen
     * @param espEncryptionAlgo
     * @param espEncryptionKey
     * @param espIv
     * @param espEncrBlockSize
     * @param espIntegrityAlgo
     * @param espIntegrityKey
     * @param espIcvLen
     * @param ipSecProtocolMode
     */
    public void setSecurityInfo(
    		BigInteger spi, 
    		int seqNbr, 
    		EnumeratedValue ahIntegrityAlgo,
    		OctetString ahIntegrityKey,
    		int ahIcvLen,
    		EnumeratedValue espEncryptionAlgo,
    		OctetString espEncryptionKey,
    		OctetString espIv,
    		int espEncrBlockSize,
    		EnumeratedValue espIntegrityAlgo, 
			OctetString espIntegrityKey,
    		int espIcvLen,
//			EnumeratedValue espCombinedModeAlgo,
//			OctetString espCombinedModeKey,
    		EnumeratedValue ipSecProtocolMode) {
    	SecurityParameterInfo sa = new SecurityParameterInfo(
    			spi,
    			seqNbr,
    			ahIntegrityAlgo,
    			ahIntegrityKey,
    			ahIcvLen, 
    			espEncryptionAlgo,
    			espEncryptionKey,
    			espIv,
    			espEncrBlockSize,
    			espIntegrityAlgo,
    			espIntegrityKey,
    			espIcvLen, 
//    			espCombinedModeAlgo,
//    			espCombinedModeKey,
    			ipSecProtocolMode
    			);
    	saMap.put(spi, sa);
    }

    /**
     * @param spi
     * @return
     */
    public boolean isSPIstored(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	return (sa!=null)?true:false;
    }
    
    /**
     * @param spi
     * @return
     */
    public int getAhIcvLen(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return sa.getAhIcvLen();
    	}
    	RB.tciTMProvided.tciError("ahIcvLen not yet set!");
    	return -1;
    }

    /**
     * @param spi
     * @return
     */
    public int getEspIcvLen(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return sa.getEspIcvLen();
    	}
    	RB.tciTMProvided.tciError("espIcvLen not yet set!");
    	return -1;
    }

    /**
     * @param spi
     * @return
     */
    public EnumeratedValue getEspIntegrityAlgo(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return sa.getEspIntegrityAlgo();
    	}
    	RB.tciTMProvided.tciError("ESP integrity algorithm not yet set!");
    	return null;
    }

    /**
     * @param spi
     * @return
     */
    public OctetstringValue getEspIntegrityKey(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return sa.getEspIntegrityKey();
    	}
    	RB.tciTMProvided.tciError("ESP integrity key not yet set!");
    	return null;
    }

    /**
     * @param spi
     * @return
     */
    public EnumeratedValue getEspEncryptionAlgo(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return sa.getEspEncryptionAlgo();
    	}
    	RB.tciTMProvided.tciError("ESP encryption algorithm not yet set!");
    	return null;
    }

    /**
     * @param spi
     * @return
     */
    public String getEspEncryptionKeyAlgo(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return getKeyAlgo(sa.getEspEncryptionAlgo().getEnum());
    	}
    	RB.tciTMProvided.tciError("ESP encryption key algorithm not yet set!");
    	return null;
    }

    /**
     * @param spi
     * @return
     */
    //TODO check if it's used
    public EnumeratedValue getIpSecProtocolMode(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return sa.getIpSecProtocolMode();
    	}
    	RB.tciTMProvided.tciError("IKE IPSEC protocol mode not yet set!");
    	return null;
    }
    
    /**
     * @param spi
     * @return
     */
    public OctetstringValue getEspEncryptionKey(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return sa.getEspEncryptionKey();
    	}
    	return null;
    }
    
    /**
     * @param spi
     * @return
     */
    public OctetString getEspIv(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return sa.getEspIv();
    	}
    	return null;
    }
    
    /**
     * @param spi
     * @return
     */
    public int getEspEncrBlockSize(BigInteger spi) {
    	SecurityParameterInfo sa = saMap.get(spi);
    	if (sa!=null) {
    		return sa.getEspEncrBlockSize();
    	}
    	return 4;
    }
    
    /**
     * @param algo
     * @return
     */
    private String getKeyAlgo(String algo) {
		if (algo.equals(ENCRYPTION_E_NULL)) {
			return "NULL";
		}
		if (algo.equals(ENCRYPTION_E_TRIPLEDES_CBC)) {
			return "TripleDES";
		}
		if (algo.equals(ENCRYPTION_E_AES_CBC) || algo.equals(ENCRYPTION_E_AES_CTR)) {
			return "AES";
		}
		if (algo.equals(ENCRYPTION_E_DES_CBC)) {
			return "DES";
		}

    	return "NULL";
    }

    /**
     * @param ikeEncryptionAlgo
     * @param ikeEncryptionKey
     * @param ikeIntegrityAlgo
     * @param ikeIntegrityKey
     */
    public void setIKESecurityInfo(
			OctetString spiInitiator,
			OctetString spiResponder,
//			IntegerValue messageID,
			EnumeratedValue ikeEncryptionAlgo,
//			OctetString ikeEncryptionKey,
			int ikeEncryptionKeyLength,
			int ikeEncrBlockSize,
			EnumeratedValue ikeIntegrityAlgo,
//			OctetString ikeIntegrityKey,
			int ikeIntegrityKeyLength,
			int ikeIcvLen,
			EnumeratedValue ikePseudoRandomFunction,
			int ikePRFKeyLength,
			EnumeratedValue diffieHellmanGroup,
			OctetString diffieHellmanPrivKey,
			OctetString diffieHellmanSharedSecret,
			RecordValue sevenSecrets) {
    	ikeSP = new IKESecurityParameterInfo(spiInitiator, spiResponder,
//				messageID.getInt(), 
				ikeEncryptionAlgo, 
//				ikeEncryptionKey,
				ikeEncryptionKeyLength,
				ikeEncrBlockSize,
				ikeIntegrityAlgo,
//				ikeIntegrityKey,
				ikeIntegrityKeyLength,
				ikeIcvLen,
				ikePseudoRandomFunction, 
				ikePRFKeyLength,
				diffieHellmanGroup,
				diffieHellmanPrivKey,
				diffieHellmanSharedSecret, 
				sevenSecrets);
    }

    /**
     * @return
     */
    public String getIkeEncryptionAlgo() {
    	if (ikeSP!=null) {
//    		return getKeyAlgo(ikeSP.getIkeEncryptionAlgo().getEnum());
    		return ikeSP.getIkeEncryptionAlgo().getEnum();
    	}
    	RB.tciTMProvided.tciError("IKE encryption algorithm not yet set!");
    	return null;
    }

    /**
     * @return
     */
    public OctetstringValue getIkeEncryptionKey(boolean initiator) {
    	if (ikeSP!=null) {
//    		return ikeSP.getIkeEncryptionKey();
    		if (initiator) {
    			return (OctetstringValue)ikeSP.getSevenSecrets().getField("sK_ei");
    		}
    		else {
    			return (OctetstringValue)ikeSP.getSevenSecrets().getField("sK_er");
    		}
    	}
    	RB.tciTMProvided.tciError("IKE encryption key not yet set!");
    	return null;
    }

    /**
     * @return
     */
    public int getIkeEncryptionKeyLength() {
    	if (ikeSP!=null) {
    		return ikeSP.getIkeEncryptionKeyLength();
    	}
    	RB.tciTMProvided.tciError("IKE encryption key length not yet set!");
    	return -1;
    }

    /**
     * @return
     */
    public int getIkeEncrBlockSize() {
    	if (ikeSP!=null) {
    		return ikeSP.getIkeEncrBlockSize();
    	}
    	RB.tciTMProvided.tciError("IKE encryption block size not yet set!");
    	return -1;
    }

    /**
     * @return
     */
    public String getIkeEncryptionKeyAlgo() {
    	if (ikeSP!=null) {
    		return getKeyAlgo(ikeSP.getIkeEncryptionAlgo().getEnum());
    	}
    	RB.tciTMProvided.tciError("IKE encryption key algorithm not yet set!");
    	return null;
    }

    /**
     * @return
     */
    public String getIkeIntegrityAlgo() {
    	if (ikeSP!=null) {
    		return ikeSP.getIkeIntegrityAlgo().getEnum();
    	}
    	RB.tciTMProvided.tciError("IKE integrity algorithm not yet set!");
    	return null;
    }

    /**
     * @return
     */
    public OctetstringValue getIkeIntegrityKey(boolean initiator) {
    	if (ikeSP!=null) {
//    		return ikeSP.getIkeIntegrityKey();
    		if (initiator) {
    			return (OctetstringValue)ikeSP.getSevenSecrets().getField("sK_ai");
    		}
    		else {
    			return (OctetstringValue)ikeSP.getSevenSecrets().getField("sK_ar");
    		}
    	}
    	RB.tciTMProvided.tciError("IKE integrity key not yet set!");
    	return null;
    }
    
    /**
     * @return
     */
    public int getIkeIntegrityKeyLength() {
    	if (ikeSP!=null) {
    		return ikeSP.getIkeIntegrityKeyLength();
    	}
    	RB.tciTMProvided.tciError("IKE authentication key length not yet set!");
    	return -1;
    }

    /**
     * @return
     */
    public int getIkeIcvLen() {
    	if (ikeSP!=null) {
    		return ikeSP.getIkeIcvLen();
    	}
    	RB.tciTMProvided.tciError("IKE ICV length not yet set!");
    	return -1;
    }

}
