/*
 * ----------------------------------------------------------------------------
 *  (C) Copyright Testing Technologies, 2001-2007.  All Rights Reserved.
 *
 *  All copies of this program, whether in whole or in part, and whether
 *  modified or not, must display this and all other embedded copyright
 *  and ownership notices in full.
 *
 *  See the file COPYRIGHT for details of redistribution and use.
 *
 *  You should have received a copy of the COPYRIGHT file along with
 *  this file; if not, write to the Testing Technologies,
 *  Rosenthaler Str. 13, 10119 Berlin, Germany.
 *
 *  TESTING TECHNOLOGIES DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
 *  SOFTWARE. IN NO EVENT SHALL TESTING TECHNOLOGIES BE LIABLE FOR ANY
 *  SPECIAL, DIRECT, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 *  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 *  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
 *  THIS SOFTWARE.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 *  EITHER EXPRESSED OR IMPLIED, INCLUDING ANY KIND OF IMPLIED OR
 *  EXPRESSED WARRANTY OF NON-INFRINGEMENT OR THE IMPLIED WARRANTIES
 *  OF MECHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * -----------------------------------------------------------------------------
 *
 *  AUTHOR:      ldte
 *  DATE:        Mar 15, 2005
 *
 *  REVISION INFO:
 *    $Revision: 1.5 $ $Date: 2005/09/13 08:47:39 $
 *    $Source: /net/ttech/cvs_root/IPv6_ETSI_STF276/javasrc/com/testingtech/ttcn/tri/JpcapRunnable.java,v $
 *
 * -----------------------------------------------------------------------------
 */


package com.testingtech.ttcn.tri;

import java.math.BigInteger;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import jpcap.IPPacket;
import jpcap.Jpcap;
import jpcap.Packet;

import org.etsi.ttcn.tri.TriComponentId;
import org.etsi.ttcn.tri.TriMessage;
import org.etsi.ttcn.tri.TriPortId;

import com.testingtech.util.HexViewer;

import de.tu_berlin.cs.uebb.muttcn.runtime.RB;

/**
 * The intention of this class is to handle all the received messages 
 * and enqueue them based on the information of the MAC address and 
 * the rcvMap to the TTCN-3 components.
 * In case of well known addresses the specific handling(e.g. enqueuing to 
 * all components) is done.
 * 
 * @author ldte
 *
 */
public class JpcapRunnable implements Runnable, TestAdapterDefinitions {
	
	private Jpcap f_jpcap = null;
	private String f_netDeviceId = "";
	
	private HashMap f_portMap = null;
	private HashMap f_rcvHashMap = null;
	
	private boolean runThread = false;
	
	private RB RB = null;
	
	/**
	 * The constructor of the class.
	 * 
	 * @param rb			The runtime used.
	 * @param jpcap			The Jpcap instance for capturing.
	 * @param netDeviceId	The net device used.
	 * @param portMap		The portMap including the component and tsiPort information.
	 * @param rcvMap		The receive hashmap inlcuding all information for the enqueuing(together with th portMap).
	 */
	public JpcapRunnable(RB rb, Jpcap jpcap, String netDeviceId, HashMap portMap, HashMap rcvMap) {
		super();
		RB = rb;
		f_jpcap = jpcap;
		f_netDeviceId = netDeviceId;
		f_portMap = portMap;
		f_rcvHashMap = rcvMap;
	}
	
	/**
	 * @param map The f_portMap to set.
	 */
	public synchronized void setPortMap(HashMap map) {
		f_portMap = map;
	}

	/**
	 * @param hashMap The f_rcvHashMap to set.
	 */
	public synchronized void setRcvHashMap(HashMap hashMap) {
		f_rcvHashMap = hashMap;
	}

	private synchronized boolean getRunThread() {
		return runThread;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		synchronized (this) {
			runThread = true;
			notifyAll();
		}

		while (getRunThread()) {
			try {
				Thread.sleep(100);
				Packet packet = f_jpcap.getPacket();

				if (packet == null) {
					continue;
				}
				
				if (getRunThread()) {
					if (RB.debug) {
						System.out.println("header: " + HexViewer.hexString(packet.header));
						System.out.println("data  : " + HexViewer.hexString(packet.data));
					}
					
					byte [] frameType = new byte[2];
					
					byte [] ipSrcAddress = null;
					byte [] ipDstAddress = null;
					byte [] ipSolNodeAddress = null;
					byte [] macSrcAddress = new byte[6];
					byte [] macDstAddress = new byte[6];
					
					byte [] ipAllNodesMca = null;
					byte [] ipAllRoutersMca = null;
					byte [] ipSolNodeMca = null;
					byte [] ipUnspecified = null;
					
					String ipPort = null;

					TriMessage trmsg = null;

					System.arraycopy(packet.header, 0, macDstAddress, 0, 6);
					System.arraycopy(packet.header, 6, macSrcAddress, 0, 6);						
					System.arraycopy(packet.header, 12, frameType, 0, 2);						

					short frameTypeValue = new BigInteger(frameType).shortValue();
					
					//Check if we handle IPv6, IPv4 or ARP - frame type
					switch (frameTypeValue) {
						case IPv6FRAMETYPE:
							ipSrcAddress = new byte[16];
							ipDstAddress = new byte[16];
							ipSolNodeAddress = new byte[13];

							System.arraycopy(packet.data,  8, ipSrcAddress, 0, 16);
							System.arraycopy(packet.data, 24, ipDstAddress, 0, 16);
							System.arraycopy(packet.data, 24, ipSolNodeAddress, 0, 13);
							
							ipAllNodesMca = IP6_ALLNODESMCA;
							ipAllRoutersMca = IP6_ALLROUTERSMCA;
							ipSolNodeMca = IP6_SOLNODEMCA;
							ipUnspecified = IP6_UNSPECIFIED;

							ipPort = IPPORT;
							
							trmsg = new TriMessageImpl(packet.data);
							break;
						case IPv4FRAMETYPE:
							//TODO Why we can't use packet.data(it's empty in fact no ipv6 is received, all stuff is in header!)
							//removing datalink + ethernet: 14 bytes
							byte [] rcvMsg = new byte[packet.header.length-14 + packet.data.length];
							//first copy the ipv4 part
							System.arraycopy(packet.header, 14, rcvMsg, 0, packet.header.length-14);
							//second try to copy the data part if present
							if (packet.data.length>0) {
								System.arraycopy(packet.data, 0, rcvMsg, packet.header.length-14, packet.data.length);
							}
							
							trmsg = new TriMessageImpl(rcvMsg);

							ipSrcAddress = ((IPPacket)packet).src_ip.getAddress();//new byte[4];
							ipDstAddress = ((IPPacket)packet).dst_ip.getAddress();//new byte[4];
							ipSolNodeAddress = new byte[13];

							ipAllNodesMca = IP4_ALLNODESMCA;
							ipAllRoutersMca = IP4_ALLROUTERSMCA;
							ipSolNodeMca = IP4_SOLNODEMCA;
							ipUnspecified = IP4_UNSPECIFIED;

							ipPort = IP4PORT;						
							break;
						case ARPFRAMETYPE:
							//TODO Why we can't use packet.data(it's empty in fact no ipv6 is received, all stuff is in header!)
							//removing datalink + ethernet: 14 bytes
							byte [] rcvArpMsg = new byte[packet.header.length-14 + packet.data.length];
							//first copy the ipv4 part
							System.arraycopy(packet.header, 14, rcvArpMsg, 0, packet.header.length-14);
							//second try to copy the data part if present
							if (packet.data.length>0) {
								System.arraycopy(packet.data, 0, rcvArpMsg, packet.header.length-14, packet.data.length);
							}
							
							trmsg = new TriMessageImpl(rcvArpMsg);

							synchronized (f_portMap) {
								for (Iterator it=f_portMap.entrySet().iterator(); it.hasNext(); ) {
									Map.Entry entry = (Map.Entry)it.next();
									TriPortId compPortId = ((TriPortId)entry.getKey());

									TriPortId tsiPortId = (TriPortId)f_portMap.get(compPortId);

									/* Do it only for ARP ports */
									if ((compPortId != null) && (compPortId.getPortName().equals(ARPPORT))) {
										if (tsiPortId != null) {
											((IPv6TestAdapter)RB.TestAdapter).enqueueMsg(tsiPortId, new TriAddressImpl(new byte[] {  }),
													compPortId.getComponent(), trmsg);
											break;
										}
										else {
											((IPv6TestAdapter)RB.TestAdapter).log(TSIPORTID_NOT_FOUND + compPortId);
											continue;
										}
									}
								}
							}
							continue;
						default:
							((IPv6TestAdapter)RB.TestAdapter).log("Frametype " + HexViewer.hexString(frameType) + " not yet supported! Ignoring!");
						continue;
					}

					/*
					 * If the received IPv6 message is a broadcast message, enqueue the message to every known component
					 */
					if (Arrays.equals(ipDstAddress, ipAllNodesMca)) { 
						synchronized (f_portMap) {
							for (Iterator it=f_portMap.entrySet().iterator(); it.hasNext(); ) {
								Map.Entry entry = (Map.Entry)it.next();
								TriPortId compPortId = ((TriPortId)entry.getKey());

								TriPortId tsiPortId = (TriPortId)f_portMap.get(compPortId);

								//    	                    		System.out.println("Enqueuing IP_ALLNODESMCA to "+ compPortId.getComponent().getComponentId());
								/* Do it only for IP ports */
								if ((compPortId != null) && !compPortId.getPortName().equals(ARPPORT)) {
									if (tsiPortId != null) {
										((IPv6TestAdapter)RB.TestAdapter).enqueueMsg(tsiPortId, new TriAddressImpl(new byte[] {  }),
												compPortId.getComponent(), trmsg);
									}
									else {
										((IPv6TestAdapter)RB.TestAdapter).log(TSIPORTID_NOT_FOUND + compPortId);
										continue;
									}
								}
							}
						}
					}
					/*
					 * If the received IPv6 message is router broadcast message, enqueue the message to every known router component
					 */
					else if (Arrays.equals(ipDstAddress, ipAllRoutersMca) || 
							(Arrays.equals(ipSolNodeAddress, ipSolNodeMca) && Arrays.equals(ipSrcAddress, ipUnspecified))
						) {
						synchronized (f_rcvHashMap) {
							synchronized (f_portMap) {
								for (Iterator it=f_portMap.entrySet().iterator(); it.hasNext(); ) {
									Map.Entry entry = (Map.Entry)it.next();
									TriComponentId componentId = ((TriPortId)entry.getKey()).getComponent();

									for (Iterator rit=f_rcvHashMap.entrySet().iterator(); rit.hasNext(); ) {
										Map.Entry rentry = (Map.Entry)rit.next();
										PTCAndRole par = ((PTCAndRole)rentry.getValue());

										if ((par != null) && (par.getComponentId().equals(componentId)) && (par.getRole().startsWith(ROUTER_ROLE))) {
											TriPortId compPortId = new TriPortIdImpl(ipPort, componentId, -1);

											TriPortId tsiPortId = (TriPortId)f_portMap.get(compPortId);

											//		                    		System.out.println("Enqueuing IP_ALLROUTERSMCA to "+ compPortId.getComponent().getComponentId());

											if (tsiPortId != null) {
												((IPv6TestAdapter)RB.TestAdapter).enqueueMsg(tsiPortId, new TriAddressImpl(new byte[] {  }),
														componentId, trmsg);
											}
											else {
												((IPv6TestAdapter)RB.TestAdapter).log(TSIPORTID_NOT_FOUND + compPortId);
												continue;
											}
											//we have found the component with role router, so don't look for anymore
											break;
										}
									}
								}
							}
						}
					}
					/*
					 * Enqueue the received IPv6 message, enqueue the message to the configured component based on the MAC address
					 */
					else {
						synchronized (f_rcvHashMap) {
							PTCAndRole par = (PTCAndRole)f_rcvHashMap.get(new BytesHash(macDstAddress));
							if (par == null) {
								((IPv6TestAdapter)RB.TestAdapter).log(TRIENQUEUEMSG_MACADDRESS_NOT_CONFIGURED + HexViewer.hexString(macDstAddress));
							}
							else {
								TriComponentId componentId = par.getComponentId();
								if (componentId != null) {
									TriPortId compPortId = new TriPortIdImpl(ipPort, componentId, -1);

									TriPortId tsiPortId = null;
									synchronized (f_portMap) {
										tsiPortId = (TriPortId)f_portMap.get(compPortId);
									}

									if (RB.debug) {
										System.out.println("Enqueuing to "+ compPortId.getComponent().getComponentId());
									}

									if (tsiPortId != null) {
										((IPv6TestAdapter)RB.TestAdapter).enqueueMsg(tsiPortId, new TriAddressImpl(new byte[] {  }),
												componentId, trmsg);
									}
									else {
										((IPv6TestAdapter)RB.TestAdapter).log(TSIPORTID_NOT_FOUND + compPortId);
										continue;
									}
								}
								else {
									((IPv6TestAdapter)RB.TestAdapter).log(TRIENQUEUEMSG_NO_COMPONENT_FOUND 
											+ ((IPv6TestAdapter)RB.TestAdapter).getIPv6AddressString(ipDstAddress) 
											+ "(" + f_netDeviceId + ")");
								}
							}
						}
					}
				}
			}
			catch (InterruptedException iex) {
				//
			}
		}
    	if (RB.debug) {
    		SimpleDateFormat df = new SimpleDateFormat("HH:mm:ss:SSS");
    		Calendar cal = Calendar.getInstance();
        	cal.setTimeInMillis(System.currentTimeMillis());
    		System.err.println(df.format(cal.getTime()) + "----------------------------------------thread end-----------------------------------" + this);
    	}
		synchronized (this) {
			notifyAll();
		}
	}
	
	/**
	 * Cancels the running receiver thread
	 */
	public synchronized void cancel() {
		while (!runThread) {
			try {
				wait();
			} catch (InterruptedException e) {
				//
			}
		}
		synchronized (this) {
			runThread = false;
		}
	}
	
}
